__all__ = [
    "census_normalize",
    "remove_zero_mvg",
    "compute_similarity2",
    "compute_similarity1",
    "compute_gcs",
    "convert_to_markov",
    "any",
    "find_nonzero",
    "FNNLSa",
    "nu",
    "eps",
    "regressed",
    "diffused",
    "compare_cytotrace",
    "cytotrace",
    "batch_cytotrace",
    "compare_cytotrace_ncores",
    "cytotrace_ncore",
    "align_diffrate",
    "plot_multiadata",
    "cumulative_boxplot",
    "run_cytotrace",
]

from time import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# Cell
import scvelo as scv
from numpy import ndarray
from scipy.stats import mannwhitneyu, rankdata

##from scanorama import correct_scanpy
scv.logging.verbosity = 3
scv.settings.presenter_view = True
scv.set_figure_params("scvelo")
np.random.seed(99)

# Cell
from scipy.sparse import csr_matrix, issparse


def census_normalize(mat, count):
    "RNA-seq census normalization to correct cell lysis"
    if issparse(mat):
        # x = mat.copy()
        # x.data = 2**x.data - 1
        # gene x cell
        x = mat.multiply(count).multiply(1.0 / mat.sum(axis=0)).tocsr()
        return x.log1p()
    else:
        x = 2**mat - 1
        # gene x cell
        x = (x * count) / x.sum(axis=0)
        return np.log2(x + 1)


# Cell
def remove_zero_mvg(mat):
    "remove cells not expressing any of the top 1000 variable genes"
    # gene x cell
    cell_count = (mat > 0).sum(axis=1)
    X_topcell = mat[cell_count >= 0.05 * mat.shape[1], :]
    var = X_topcell.var(axis=1)
    mean = X_topcell.mean(axis=1)
    disp = var / mean
    return X_topcell[
        disp >= (np.sort(disp)[::-1])[:1000][-1], :
    ]  # in case less than 1000 genes


# Cell
def compute_similarity2(O: ndarray, P: ndarray) -> ndarray:
    "Compute pearson correlation between two matrices O and P using einstein summation"
    (n, t) = O.shape  # n traces of t samples
    (n_bis, m) = P.shape  # n predictions for each of m candidates

    DO = O - (
        np.einsum("nt->t", O, optimize="optimal") / np.double(n)
    )  # compute O - mean(O)
    DP = P - (
        np.einsum("nm->m", P, optimize="optimal") / np.double(n)
    )  # compute P - mean(P)

    cov = np.einsum("nm,nt->mt", DP, DO, optimize="optimal")

    varP = np.einsum("nm,nm->m", DP, DP, optimize="optimal")
    varO = np.einsum("nt,nt->t", DO, DO, optimize="optimal")
    tmp = np.einsum("m,t->mt", varP, varO, optimize="optimal")
    return cov / np.sqrt(tmp)


# Cell
# adapt from https://github.com/ikizhvatov/efficient-columnwise-correlation/blob/master/columnwise_corrcoef_perf.py
def compute_similarity1(A):
    "Compute pairwise correlation of all columns in matrices A"
    n, t = A.shape  # genes x samples
    DO = A - (
        np.einsum("nt->t", A, optimize="optimal") / np.double(n)
    )  # compute O - mean(O)
    cov = np.einsum("nm,nt->mt", DO, DO, optimize="optimal")
    varP = np.einsum("nt,nt->t", DO, DO, optimize="optimal")
    return cov / np.sqrt(np.einsum("m,t->mt", varP, varP, optimize="optimal"))


# Cell
def compute_gcs(mat, count, top_n_genes=200):
    "Compute gene set enrichment scores by correlating gene count and gene expression"
    corrs = compute_similarity2(mat.T, count.reshape(-1, 1))[0, :]
    # avoid nan, put all nan to -1
    corrs[np.isnan(corrs)] = -1
    gcs = mat[np.argsort(corrs)[::-1][:top_n_genes], :].mean(axis=0)
    return gcs


# Cell
def convert_to_markov(sim):
    """Convert the Pearson correlation to Markov matrix

    TODO: use velocity graph to replace this markov matrix
    """
    Ds = np.copy(sim)

    cutoff = np.mean(Ds)
    np.fill_diagonal(Ds, 0)

    Ds[Ds <= 0] = 0
    Ds[Ds <= cutoff] = 0

    zero_rows = Ds.sum(axis=1) == 0
    zero_cols = Ds.sum(axis=0) == 0

    Ds[~zero_rows, :] = (
        Ds[~zero_rows, :].T / Ds[~zero_rows, :].sum(axis=1)
    ).T  # tips
    Ds[:, zero_cols] = 0
    return Ds


# Cell
# http://xrm.phys.northwestern.edu/research/pdf_papers/1997/bro_chemometrics_1997.pdf
# matlab reference: https://www.mathworks.com/matlabcentral/mlc-downloads/downloads/submissions/3388/versions/1/previews/fnnls.m/index.html

from numpy import (
    abs,
    arange,
    argmax,
    finfo,
    float64,
    int64,
    min,
    newaxis,
    nonzero,
    sum,
    zeros,
)
from scipy.linalg import solve

nu = newaxis
import numpy as np

# machine epsilon
eps = finfo(float64).eps


def any(a):
    # assuming a vector, a
    larger_than_zero = sum(a > 0)
    if larger_than_zero:
        return True
    else:
        return False


def find_nonzero(a):
    # returns indices of nonzero elements in a
    return nonzero(a)[0]


def FNNLSa(XtX, Xty, tol=None):
    """Faster NNLS imported from https://github.com/delnatan/FNNLSa
    A fast non-negativity-constrained least squares algorithm. Journal of chemometrics
    """

    if tol is None:
        tol = eps

    M, N = XtX.shape
    # initialize passive set, P. Indices where coefficient is >0
    P = zeros(N, dtype=int64)
    # and active set. Indices where coefficient is <=0
    Z = arange(N) + 1
    # working active set
    ZZ = arange(N) + 1
    # initial solution vector, x
    x = zeros(N, dtype=float64)
    # weight vector
    w = Xty - XtX @ x
    # iteration counts and parameter
    it = 0
    itmax = 30 * N
    # MAIN LOOP
    # continue as long as there are indices within the active set Z
    # or elements in inner loop active set is larger than 'tolerance'
    piter = 0
    while any(Z) and any(w[ZZ - 1] > tol):
        piter += 1
        t = argmax(w[ZZ - 1]) + 1  # find largest weight
        t = ZZ[t - 1]
        P[t - 1] = t  # move to passive set
        Z[t - 1] = 0  # remove from active set
        PP = find_nonzero(P) + 1
        ZZ = find_nonzero(Z) + 1
        NZZ = ZZ.shape

        # compute trial solution, s
        s = zeros(N, dtype=float64)

        if len(PP) == 1:
            s[PP - 1] = Xty[PP - 1] / XtX[PP - 1, PP - 1]
        else:
            s[PP - 1] = solve(XtX[PP - 1, PP[:, nu] - 1], Xty[PP - 1])
        s[ZZ - 1] = 0.0  # set active coefficients to 0

        while any(s[PP - 1] <= tol) and it < itmax:
            it = it + 1
            QQ = find_nonzero((s <= tol) * P) + 1
            alpha = min(x[QQ - 1] / (x[QQ - 1] - s[QQ - 1]))
            x = x + alpha * (s - x)
            ij = find_nonzero((abs(x) < tol) * (P != 0)) + 1
            Z[ij - 1] = ij
            P[ij - 1] = 0
            PP = find_nonzero(P) + 1
            ZZ = find_nonzero(Z) + 1
            if len(PP) == 1:
                s[PP - 1] = Xty[PP - 1] / XtX[PP - 1, PP - 1]
            else:
                s[PP - 1] = solve(XtX[PP - 1, PP[:, nu] - 1], Xty[PP - 1])
            s[ZZ - 1] = 0.0
        # assign current solution, s, to x
        x = s
        # recompute weights
        w = Xty - XtX @ x
    return x, w


# Cell
def regressed(markov, gcs, solver="fnnls"):
    """solve markov @ weight = gcs problems,

    solver: fnnls (default) is faster in larger dataset, e.g., above 20,000 cells
            nnls is faster in smaller dataset, e.g. less than 5,000 cells
    """
    print(solver)
    if solver == "fnnls":
        coef, err = FNNLSa(markov.T @ markov, markov.T @ gcs)
    elif (
        solver == "jfnnls"
    ):  # pip install fnnls # https://github.com/jvendrow/fnnls
        from fnnls import fnnls

        coef, err = fnnls(markov, gcs)
    elif solver == "nnls":
        from scipy.optimize import nnls

        print((markov > 0).sum())
        coef, err = nnls(markov, gcs)  # from 1987...
    elif solver == "lsq_linear":
        from scipy import sparse
        from scipy.optimize import lsq_linear

        # slow as well
        sol = lsq_linear(
            sparse.csr_matrix(markov),
            gcs,
            bounds=(0, np.inf),
            lsmr_tol="auto",
            verbose=1,
        )  # from 1997...
        coef = sol.x
    elif solver == "lasso":
        from scipy import sparse
        from sklearn.linear_model import Lasso

        lasso = Lasso(
            alpha=1e-7, max_iter=100000, fit_intercept=False, positive=True
        )
        den = (markov > 0).sum() / np.prod(markov.shape)
        print(den)
        if den >= 0.5:
            lasso.fit(markov, gcs)
        else:
            # lasso.fit(sparse.csr_matrix(markov), gcs)
            lasso.fit(sparse.coo_matrix(markov), gcs)
        coef = lasso.coef_
    elif solver == "nnlsgd":  # somehow not consistent with nnls and fnnls...
        coef = NNLS_GD(markov.T @ markov, markov.T @ gcs)
    elif solver == None:  # do not fit
        return gcs
    else:  # tnt-nn
        raise NotImplementedError
    # faster nnls
    return np.dot(markov, coef)


# Cell
def diffused(markov, gcs, ALPHA=0.9):
    """Compute diffusion process"""
    v_prev = np.copy(gcs)
    v_curr = np.copy(gcs)

    for i in range(10000):
        v_prev = np.copy(v_curr)
        v_curr = ALPHA * (markov.dot(v_curr)) + (1 - ALPHA) * gcs

        diff = np.mean(np.abs(v_curr - v_prev))
        if diff <= 1e-6:
            break
    return v_curr


# Cell
from scipy.sparse import issparse


def compare_cytotrace(
    adata,
    layer="all",
    cell_count=10,
    condition="age",
    solver="nnls",
    is_normalized=False,
    n_cores=4,
    top_n_genes=200,
):
    "Main interface of cytotrace reimplementation used for single dataset with multiple conditions"
    # condition common steps

    proc_time = time()
    if not is_normalized:
        if layer == "all":
            X = (adata.layers["spliced"] + adata.layers["unspliced"]).toarray()
        else:
            try:
                X = (
                    adata.layers[layer].toarray()
                    if issparse(adata.layers[layer])
                    else adata.layers[layer]
                )
            except:
                X = adata.X.toarray() if issparse(adata.X) else adata.X

        cells_selected = np.arange(X.shape[0])
        genes_selected = np.arange(X.shape[1])
        pgenes = (pd.isnull((X > 0).sum(axis=0))) | (X.var(axis=0) == 0)
        # cell x gene
        X = X[:, ~pgenes]
        X = (X.T / X.sum(axis=1)) * 1e6

        pqcells = (pd.isnull((X > 0).sum(axis=0))) | (
            (X > 0).sum(axis=0) <= cell_count
        )
        X = X[:, ~pqcells]

        genes_selected = genes_selected[~pgenes]
        cells_selected = cells_selected[~pqcells]

        X = np.log2(X + 1)
        counts = (X > 0).sum(axis=0)
        mat2 = census_normalize(X, counts)

        mvg = remove_zero_mvg(mat2)
        cells_selected = cells_selected[mvg.sum(axis=0) != 0]

        mat2 = mat2[:, mvg.sum(axis=0) != 0]
        counts = counts[mvg.sum(axis=0) != 0]

        # joint two conditions to determine gcs feature
        # using the same set of genes
        gcs = compute_gcs(mat2, counts, top_n_genes=top_n_genes)
        adata.uns["gcs"] = gcs
        adata.uns["cells_selected"] = cells_selected
        adata.uns["counts"] = counts
        adata.uns["mat2"] = mat2
        adata.uns["mvg"] = mvg
        adata.uns["genes_selected"] = genes_selected
    else:
        cells_selected = adata.uns["cells_selected"]
        counts = adata.uns["counts"]
        mat2 = adata.uns["mat2"]
        mvg = adata.uns["mvg"]
        genes_selected = adata.uns["genes_selected"]
        gcs = adata.uns["gcs"]

    # for indepent gcs feature
    # using different set of genes is very variable
    # gcs = np.zeros(len(cells_selected))

    # condition-specific steps
    for cond in np.unique(adata.obs.loc[:, condition].values[cells_selected]):
        selection = adata.obs.loc[:, condition].values[cells_selected] == cond

        gcs_cond = gcs[selection]
        counts_cond = counts[selection]
        mat2_cond = mat2[:, selection]

        # compute gcs feature independently for two conditions
        # this would select different gene set, and make the
        # cytotrace score incomparable
        # gcs_cond = compute_gcs(mat2_cond, counts_cond)

        mvg_cond = mvg[:, selection]

        trans_time = time()
        print("preprocessing: %s" % (trans_time - proc_time))
        # get transition matrix
        corr_cond = compute_similarity1(mvg_cond)
        markov_cond = convert_to_markov(corr_cond)

        gc_time = time()
        print("markov: %s" % (gc_time - trans_time))

        # get Gene count signature
        gcs_cond = regressed(markov_cond, gcs_cond, solver=solver)
        gcs_cond = diffused(markov_cond, gcs_cond)
        gcs[selection] = gcs_cond

        # condition specific cytotrace correlated gene sets
        scores_cond = rankdata(gcs_cond) / gcs_cond.shape[0]
        score_time = time()
        print("score time: %s" % (score_time - gc_time))

        corrs_cond = compute_similarity2(
            mat2_cond.T, scores_cond.reshape(-1, 1)
        )[0, :]
        final_time = time()
        print("final time: %s" % (final_time - score_time))
        adata.var["cytotrace_corrs_%s" % cond] = None
        adata.var.iloc[
            genes_selected,
            adata.var.columns.get_loc("cytotrace_corrs_%s" % cond),
        ] = corrs_cond

    adata.obs["gcs"] = None
    adata.obs["cytotrace"] = None
    adata.obs["counts"] = None
    adata.obs.iloc[
        cells_selected, adata.obs.columns.get_loc("gcs")
    ] = gcs  # used for re-rank when integrating multiple-samples
    adata.obs.iloc[cells_selected, adata.obs.columns.get_loc("cytotrace")] = (
        rankdata(gcs) / gcs.shape[0]
    )  # cytotrace scores
    adata.obs.iloc[
        cells_selected, adata.obs.columns.get_loc("counts")
    ] = counts  # used for re-rank when integrating multiple-samples


def cytotrace_sparse(
    adata,
    layer="raw",
    cell_count=20,
    solver="nnls",
    top_n_features=200,
    skip_regress=False,
):
    "optimized version"
    proc_time = time()
    if not issparse(adata.layers[layer]):
        raise NotImplementedError
    else:
        X = adata.layers[layer]

    cells_selected = np.arange(X.shape[0])
    features_selected = np.arange(X.shape[1])

    n_cells = X.shape[0]
    feature_mean = (X.sum(0) / n_cells).A1
    feature_mean_sq = (X.multiply(X).sum(0) / n_cells).A1
    feature_var = (feature_mean_sq - feature_mean**2) * (
        n_cells / (n_cells - 1)
    )
    non_zero_features = (X > 0).sum(axis=0).A1
    pfeatures = np.isnan(non_zero_features) | (feature_var == 0)

    # cell x feature, filter feature
    X = X[:, ~pfeatures]
    features_selected = features_selected[~pfeatures]

    # gene x cell
    X = X.T
    X = X.multiply(1.0 / csr_matrix.sum(X, axis=0)) * 1e6
    X = X.tocsr()

    # filter cell, feature x cell
    feature_count_per_cell = (X > 0).sum(axis=0).A1
    pcells = np.isnan(feature_count_per_cell) | (
        feature_count_per_cell < cell_count
    )

    # feature x cell
    X = X[:, ~pcells]
    cells_selected = cells_selected[~pcells]
    feature_count_per_cell = feature_count_per_cell[~pcells]

    # census normalize, feature x cell
    census_X = (
        X.multiply(feature_count_per_cell)
        .multiply(1.0 / csr_matrix.sum(X, axis=0))
        .log1p()
        .tocsr()
    )

    # top variable features, feature x cell
    cell_count_per_feature = (census_X > 0).sum(axis=1).A1
    census_X_topcell = census_X[
        cell_count_per_feature >= 0.05 * census_X.shape[1], :
    ].tocsr()
    n_cells = census_X_topcell.shape[1]
    feature_mean = (census_X_topcell.sum(1) / n_cells).A1
    feature_mean_sq = (
        census_X_topcell.multiply(census_X_topcell).sum(1) / n_cells
    ).A1
    feature_var = (feature_mean_sq - feature_mean**2) * (
        n_cells / (n_cells - 1)
    )
    disp = feature_var / feature_mean
    mvg = census_X_topcell[
        disp >= (np.sort(disp)[::-1])[:1000][-1], :
    ]  # in case less than 1000 features

    # top 1000 variable features for markov matrix
    selection = (mvg.sum(axis=0) != 0).A1
    mvg = mvg[:, selection]
    corr = compute_similarity1(mvg.A)
    markov = convert_to_markov(corr)

    # filter census output with nonzero cells in top variable features matrix
    cells_selected = cells_selected[selection]
    census_X = census_X[:, selection]
    feature_count_per_cell = feature_count_per_cell[selection]

    # calculate GC scores
    corrs = compute_similarity2(
        census_X.T.A, feature_count_per_cell.reshape(-1, 1)
    )[0, :]
    # avoid nan, put all nan to -1
    corrs[np.isnan(corrs)] = -1
    gcs = census_X[np.argsort(corrs)[::-1][:top_n_features], :].mean(axis=0).A1

    if not skip_regress:
        print("regress...")
        # from scipy.optimize import lsq_linear
        # # slow as well
        # sol = lsq_linear(csr_matrix(markov), gcs, bounds=(0, np.inf), lsmr_tol='auto', verbose=1)    # from 1997...
        # coef = sol.x

        from scipy.optimize import nnls

        coef, err = nnls(markov, gcs)  # from 1987...
        gcs = np.dot(markov, coef)

    print(markov.shape, gcs.shape)
    gcs = diffused(markov, gcs)
    rank = rankdata(gcs)
    scores = rank / gcs.shape[0]

    print(gcs)
    adata.obs["gcs"] = np.nan
    adata.obs.iloc[cells_selected, adata.obs.columns.get_loc("gcs")] = gcs

    cytoGenes = compute_similarity2(census_X.T.A, scores.reshape(-1, 1))[0, :]

    adata.obs["cytotrace"] = np.nan
    adata.obs.iloc[
        cells_selected, adata.obs.columns.get_loc("cytotrace")
    ] = scores

    adata.obs["counts"] = np.nan
    adata.obs.iloc[cells_selected, adata.obs.columns.get_loc("counts")] = gcs

    adata.var["cytotrace"] = False
    adata.var.iloc[
        features_selected, adata.var.columns.get_loc("cytotrace")
    ] = True

    adata.var["cytotrace_corrs"] = np.nan
    adata.var.iloc[
        features_selected, adata.var.columns.get_loc("cytotrace_corrs")
    ] = np.array(corrs, dtype=np.float32)
    return {
        "CytoTRACE": scores,
        "CytoTRACErank": rank,
        "GCS": gcs,
        "Counts": feature_count_per_cell,
        "exprMatrix": census_X,
        "cytoGenes": cytoGenes,
        "gcsGenes": corrs,
        "filteredCells": np.setdiff1d(
            np.arange(adata.shape[0]), cells_selected
        ),
    }


# Cell
def cytotrace(
    adata, layer="all", cell_count=10, solver="nnls", top_n_genes=200
):
    "Main interface of cytotrace reimplementation used for single dataset with one condition"
    proc_time = time()
    if layer == "all":
        X = (adata.layers["spliced"] + adata.layers["unspliced"]).toarray()
    #        X = (adata.layers['spliced'] + adata.layers['unspliced'])
    else:
        try:
            X = (
                adata.layers[layer].toarray()
                if issparse(adata.layers[layer])
                else adata.layers[layer]
            )
            # X = adata.layers[layer] if issparse(adata.layers[layer]) else adata.layers[layer]
        except:
            X = adata.X.toarray() if issparse(adata.X) else adata.X
            # X = adata.X if issparse(adata.X) else adata.X

    cells_selected = np.arange(X.shape[0])
    genes_selected = np.arange(X.shape[1])

    pgenes = (pd.isnull((X > 0).sum(axis=0))) | (X.var(axis=0) == 0)
    # cell x gene
    X = X[:, ~pgenes]
    # gene x cell
    X = (X.T / X.sum(axis=1)) * 1e6

    pqcells = (pd.isnull((X > 0).sum(axis=0))) | (
        (X > 0).sum(axis=0) <= cell_count
    )
    X = X[:, ~pqcells]

    genes_selected = genes_selected[~pgenes]
    cells_selected = cells_selected[~pqcells]

    X = np.log2(X + 1)
    counts = (X > 0).sum(axis=0)
    mat2 = census_normalize(X, counts)

    mvg = remove_zero_mvg(mat2)
    selection = mvg.sum(axis=0) != 0
    mvg = mvg[:, selection]
    mat2 = mat2[:, selection]
    counts = counts[selection]

    cells_selected = cells_selected[selection]

    trans_time = time()
    print("processing: %s" % (trans_time - proc_time))

    # get transition matrix
    corr = compute_similarity1(mvg)
    markov = convert_to_markov(corr)

    gc_time = time()
    print("markov: %s" % (gc_time - trans_time))

    # get Gene count signature
    # gcs = compute_gcs(mat2, counts)
    # gcs = compute_gcs(mat2, counts, top_n_genes=top_n_genes)

    corrs = compute_similarity2(mat2.T, counts.reshape(-1, 1))[0, :]
    # avoid nan, put all nan to -1
    corrs[np.isnan(corrs)] = -1
    gcs = mat2[np.argsort(corrs)[::-1][:top_n_genes], :].mean(axis=0)

    gcs_time = time()
    print("gcs: %s" % (gcs_time - gc_time))

    gcs = regressed(markov, gcs, solver=solver)
    regress_time = time()
    print("regression: %s" % (regress_time - gcs_time))

    gcs = diffused(markov, gcs)
    rank = rankdata(gcs)
    scores = rankdata(gcs) / gcs.shape[0]
    score_time = time()
    print("diffusion:%s " % (score_time - regress_time))

    cyto_corrs = compute_similarity2(mat2.T, scores.reshape(-1, 1))[0, :]
    gene_time = time()
    print("genes:%s" % (gene_time - score_time))

    adata.obs["gcs"] = None
    adata.obs.iloc[cells_selected, adata.obs.columns.get_loc("gcs")] = gcs

    adata.obs["cytotrace"] = None
    adata.obs.iloc[
        cells_selected, adata.obs.columns.get_loc("cytotrace")
    ] = scores

    adata.obs["counts"] = None
    adata.obs.iloc[cells_selected, adata.obs.columns.get_loc("counts")] = counts

    adata.var["cytotrace"] = None
    adata.var.iloc[
        genes_selected, adata.var.columns.get_loc("cytotrace")
    ] = True

    adata.var["cytotrace_corrs"] = None
    adata.var.iloc[
        genes_selected, adata.var.columns.get_loc("cytotrace_corrs")
    ] = cyto_corrs
    return {
        "CytoTRACE": scores,
        "CytoTRACErank": rank,
        "GCS": gcs,
        "Counts": counts,
        "exprMatrix": mat2,
        "cytoGenes": cyto_corrs,
        "gcsGenes": corrs,
        "filteredCells": np.setdiff1d(
            np.arange(adata.shape[0]), cells_selected
        ),
    }


def visualize(adata, metrics, name="test"):
    import scanpy as sc
    import seaborn as sns

    fig, ax = plt.subplots(2)
    fig.set_size_inches(8, 6)
    sc.pl.umap(adata, color="cytotrace", show=False, ax=ax[0])
    order = (
        adata.obs.groupby("clusters").median().sort_values(["cytotrace"]).index
    )
    bplot = sns.boxplot(
        x="clusters",
        y="cytotrace",
        data=adata.obs,
        order=order,
        width=0.35,
        ax=ax[1],
    )
    ax[1].set_xticklabels(order, rotation=40, ha="right")
    fig.savefig(f"{name}_figure.pdf")


# Cell
def batch_cytotrace(mvg_batch, gcs_batch, solver="jfnnls"):
    corr_batch = compute_similarity1(mvg_batch)
    markov_batch = convert_to_markov(corr_batch)
    gcs_time = time()
    gcs_batch = regressed(markov_batch, gcs_batch, solver=solver)
    regress_time = time()
    print("regression: %s" % (regress_time - gcs_time))
    print(markov_batch.shape, gcs_batch.shape)
    gcs_batch = diffused(markov_batch, gcs_batch)
    return gcs_batch


# Cell
from scipy.sparse import issparse


def compare_cytotrace_ncores(
    adata,
    layer="all",
    cell_count=10,
    condition="age",
    solver="nnls",
    is_normalized=False,
    ncores=4,
    batch_cell=2000,
):
    "Main interface of cytotrace reimplementation used for single dataset with multiple conditions"
    # condition common steps

    proc_time = time()
    if not is_normalized:
        if layer == "all":
            X = (
                adata.layers["spliced"] + adata.layers["unspliced"]
            )  # .toarray()
        else:
            try:
                X = (
                    adata.layers[layer]
                    if issparse(adata.layers[layer])
                    else adata.layers[layer]
                )
            except:
                X = adata.X if issparse(adata.X) else adata.X

        cells_selected = np.arange(X.shape[0])
        genes_selected = np.arange(X.shape[1])

        from sklearn.utils import sparsefuncs_fast

        mean, var = sparsefuncs_fast.csr_mean_variance_axis0(X)

        pgenes = (np.isnan(np.array((X > 0).sum(axis=0))[0])) | (var == 0)

        # cell x gene
        X = X[:, ~pgenes]
        X = (X.T).multiply(1.0 / csr_matrix.sum(X.T, axis=0)).tocsr()
        pqcells = np.array(
            (np.isnan((X > 0).sum(axis=0)))
            | ((X > 0).sum(axis=0) <= cell_count)
        )[0]
        X = X[:, ~pqcells]

        genes_selected = genes_selected[~pgenes]
        cells_selected = cells_selected[~pqcells]

        counts = np.array((X > 0).sum(axis=0))[0]
        mat2 = census_normalize(X, counts).toarray()
        mvg = remove_zero_mvg(mat2)

        # sel = np.array(mvg.sum(axis=0))[0]!=0
        sel = mvg.sum(axis=0) != 0
        mat2 = mat2[:, sel]
        counts = counts[sel]
        cells_selected = cells_selected[sel]

        # joint two conditions to determine gcs feature
        # using the same set of genes
        gcs = compute_gcs(mat2, counts)
        adata.uns["gcs"] = gcs
        adata.uns["cells_selected"] = cells_selected
        adata.uns["counts"] = counts
        adata.uns["mat2"] = mat2
        adata.uns["mvg"] = mvg
        adata.uns["genes_selected"] = genes_selected
    else:
        cells_selected = adata.uns["cells_selected"]
        counts = adata.uns["counts"]
        mat2 = adata.uns["mat2"]
        mvg = adata.uns["mvg"]
        genes_selected = adata.uns["genes_selected"]
        gcs = adata.uns["gcs"]

    # for indepent gcs feature
    # using different set of genes is very variable
    # gcs = np.zeros(len(cells_selected))

    # parallel part
    import multiprocessing as mp

    # https://stackoverflow.com/questions/8533318/multiprocessing-pool-when-to-use-apply-apply-async-or-map
    # condition-specific steps
    pool = mp.Pool(ncores)

    for cond in np.unique(adata.obs.loc[:, condition].values[cells_selected]):
        print(cond)
        selection = adata.obs.loc[:, condition].values[cells_selected] == cond

        gcs_cond = gcs[selection]
        counts_cond = counts[selection]
        mat2_cond = mat2[:, selection]
        mvg_cond = mvg[:, selection]

        mvg_gcs_batches = []
        for i in range(0, gcs_cond.shape[0], batch_cell):
            mvg_batch = mvg_cond[:, i : (i + batch_cell)].copy()
            gcs_batch = gcs_cond[i : i + batch_cell].copy()
            print(mvg_batch.shape, gcs_batch.shape)
            mvg_gcs_batches.append([mvg_batch, gcs_batch, solver])

        # https://stackoverflow.com/questions/8533318/multiprocessing-pool-when-to-use-apply-apply-async-or-map
        # http://jennguyen1.github.io/nhuyhoa/software/Parallel-Processing.html
        result = pool.starmap_async(batch_cytotrace, mvg_gcs_batches)
        result.wait()

        gcs_batches = result.get()
        # gcs_batches = result.get()
        gcs_cond = np.hstack(gcs_batches)

        trans_time = time()
        print("preprocessing: %s" % (trans_time - proc_time))

        gc_time = time()
        print("markov: %s" % (gc_time - trans_time))
        gcs[selection] = gcs_cond

        # condition specific cytotrace correlated gene sets
        scores_cond = rankdata(gcs_cond) / gcs_cond.shape[0]
        score_time = time()
        print("score time: %s" % (score_time - gc_time))

        corrs_cond = compute_similarity2(
            mat2_cond.T, scores_cond.reshape(-1, 1)
        )[0, :]
        final_time = time()
        print("final time: %s" % (final_time - score_time))
        adata.var["cytotrace_corrs_%s" % cond] = None
        adata.var.iloc[
            genes_selected,
            adata.var.columns.get_loc("cytotrace_corrs_%s" % cond),
        ] = corrs_cond

    pool.close()
    pool.join()

    adata.obs["gcs"] = None
    adata.obs["cytotrace"] = None
    adata.obs["counts"] = None
    adata.obs.iloc[
        cells_selected, adata.obs.columns.get_loc("gcs")
    ] = gcs  # used for re-rank when integrating multiple-samples
    adata.obs.iloc[cells_selected, adata.obs.columns.get_loc("cytotrace")] = (
        rankdata(gcs) / gcs.shape[0]
    )  # cytotrace scores
    adata.obs.iloc[
        cells_selected, adata.obs.columns.get_loc("counts")
    ] = counts  # used for re-rank when integrating multiple-samples


# Cell


def cytotrace_ncore(
    adata,
    layer="all",
    cell_count=10,
    solver="nnls",
    ncores=4,
    batch_cell=3000,
    shuffle=3,
):
    "optimized version"
    proc_time = time()
    if layer == "all":
        X = adata.layers["spliced"] + adata.layers["unspliced"]
    else:
        try:
            X = (
                adata.layers[layer]
                if issparse(adata.layers[layer])
                else adata.layers[layer]
            )
        except:
            X = adata.X if issparse(adata.X) else adata.X

    cells_selected = np.arange(X.shape[0])
    genes_selected = np.arange(X.shape[1])

    from sklearn.utils import sparsefuncs_fast

    mean, var = sparsefuncs_fast.csr_mean_variance_axis0(X)

    pgenes = (np.isnan(np.array((X > 0).sum(axis=0))[0])) | (var == 0)
    # pgenes = np.isnan(mean * adata.shape[0]) | (var == 0)

    # cell x gene
    X = X[:, ~pgenes]
    # gene x cell
    # X = (X.T / np.array(X.sum(axis=1))[0]) * 1e6
    # sparse operation
    X = (X.T).multiply(1.0 / csr_matrix.sum(X.T, axis=0)).tocsr()

    pqcells = np.array(
        (np.isnan((X > 0).sum(axis=0))) | ((X > 0).sum(axis=0) <= cell_count)
    )[0]
    X = X[:, ~pqcells]

    genes_selected = genes_selected[~pgenes]
    cells_selected = cells_selected[~pqcells]

    counts = np.array((X > 0).sum(axis=0))[0]
    mat2 = census_normalize(X, counts).toarray()

    # cell_count = np.array((mat2 > 0).sum(axis=1))[:, 0]
    # X_topcell = mat2[cell_count >= 0.05 * mat2.shape[1], :]

    # mean, var = sparsefuncs_fast.csr_mean_variance_axis0(X_topcell)
    # disp = var / mean
    # mvg = X_topcell[disp >= (np.sort(disp)[::-1])[:1000][-1], :]
    mvg = remove_zero_mvg(mat2)

    # sel = np.array(mvg.sum(axis=0))[0]!=0
    sel = mvg.sum(axis=0) != 0
    mat2 = mat2[:, sel]
    counts = counts[sel]
    cells_selected = cells_selected[sel]

    # mat2 = mat2.toarray()
    # mvg = mvg.toarray()
    print(mvg.shape, mat2.shape)

    trans_time = time()
    print("processing: %s" % (trans_time - proc_time))

    gcs = compute_gcs(mat2, counts)

    # get transition matrix
    # parallel part
    import multiprocessing as mp

    pool = mp.Pool(ncores)

    # https://stackoverflow.com/questions/8533318/multiprocessing-pool-when-to-use-apply-apply-async-or-map
    mvg_gcs_batches = []
    for i in range(0, adata.shape[0], batch_cell):
        mvg_batch = mvg[:, i : (i + batch_cell)].copy()
        gcs_batch = gcs[i : i + batch_cell].copy()
        mvg_gcs_batches.append([mvg_batch, gcs_batch, solver])

    markov_diffusion_time = time()

    # https://stackoverflow.com/questions/8533318/multiprocessing-pool-when-to-use-apply-apply-async-or-map
    result = pool.starmap_async(batch_cytotrace, mvg_gcs_batches)
    pool.close()
    pool.join()
    score_time = time()
    print(score_time - markov_diffusion_time)
    gcs_batches = result.get()
    gcs_batches = np.hstack(gcs_batches)

    scores = rankdata(gcs_batches) / gcs_batches.shape[0]
    corrs = compute_similarity2(mat2.T, scores.reshape(-1, 1))[0, :]

    gene_time = time()
    print("genes:%s" % (gene_time - score_time))
    adata.obs["gcs"] = None
    adata.obs.iloc[cells_selected, adata.obs.columns.get_loc("gcs")] = gcs

    adata.obs["cytotrace"] = None
    adata.obs.iloc[
        cells_selected, adata.obs.columns.get_loc("cytotrace")
    ] = scores

    adata.obs["counts"] = None
    adata.obs.iloc[cells_selected, adata.obs.columns.get_loc("counts")] = counts

    adata.var["cytotrace"] = None
    adata.var.iloc[
        genes_selected, adata.var.columns.get_loc("cytotrace")
    ] = True

    adata.var["cytotrace_corrs"] = None
    adata.var.iloc[
        genes_selected, adata.var.columns.get_loc("cytotrace_corrs")
    ] = corrs


# Cell
def align_diffrate(
    adatas, labels, field="condition", type="A", outfield="cytotrace", ax=None
):
    "this is used for differentiation rate comparison across samples"
    scores = [
        adatas[0].obs.loc[:, outfield][adatas[0].obs.loc[:, field] == type],
        adatas[1].obs.loc[:, outfield][adatas[1].obs.loc[:, field] == type],
    ]
    if outfield in ["latent_time", "velocity_pseudotime"]:
        pvalg = mannwhitneyu(scores[0], scores[1], alternative="greater")[1]
        pvall = mannwhitneyu(scores[0], scores[1], alternative="less")[1]
    else:
        pvalg = mannwhitneyu(
            1 - scores[0], 1 - scores[1], alternative="greater"
        )[1]
        pvall = mannwhitneyu(1 - scores[0], 1 - scores[1], alternative="less")[
            1
        ]

    for s, l in zip(scores, labels):
        if outfield in ["latent_time", "velocity_pseudotime"]:
            s = np.sort(s)
        else:
            if outfield != "cytotrace":
                s = np.sort(-s)
            else:
                s = np.sort(1 - s)
        ax.step(
            np.concatenate([s, s[[-1]]]),
            np.arange(s.size + 1) / s.size,
            label=l,
        )
        ax.scatter(
            np.concatenate([s, s[[-1]]]), np.arange(s.size + 1) / s.size, s=5
        )
        if outfield == "cytotrace":
            ax.set_xlabel("Differentiation level")
        else:
            ax.set_xlabel(outfield)
        ax.set_ylabel("Cumulative proportion of cells")
        ax.set_title(type)
    ax.text(
        np.percentile(s, 0.6),
        0.4,
        f"{labels[0]} > {labels[1]}: {pvalg:.3E}",
    )
    ax.text(
        np.percentile(s, 0.6),
        0.48,
        f"{labels[0]} < {labels[1]}: {pvall:.3E}",
    )
    ax.legend()


# Cell
def plot_multiadata(adatas):
    for a in adatas:
        print(a.obs.age[0])
        fig, ax = plt.subplots(1, 5)
        fig.set_size_inches(32, 9)
        scv.pl.velocity_embedding_grid(
            a,
            scale=0.2,
            color="louvain",
            show=False,
            basis="pca",
            legend_loc="on data",
            ax=ax[0],
            title=a.obs.age[0],
        )
        scv.pl.scatter(
            a,
            color="velocity_pseudotime",
            basis="pca",
            size=80,
            color_map="gnuplot",
            ax=ax[1],
            show=False,
            title=a.obs.age[0],
        )
        scv.pl.scatter(
            a,
            color="latent_time",
            basis="pca",
            size=80,
            color_map="gnuplot",
            ax=ax[2],
            show=False,
            title=a.obs.age[0],
        )
        scv.pl.scatter(
            a,
            color="end_points",
            basis="pca",
            size=80,
            color_map="gnuplot",
            ax=ax[3],
            show=False,
            title=a.obs.age[0],
        )
        scv.pl.scatter(
            a,
            color="root_cells",
            basis="pca",
            size=80,
            color_map="gnuplot",
            ax=ax[4],
            show=False,
            title=a.obs.age[0],
        )


# Cell
# https://stackoverflow.com/questions/15408371/cumulative-distribution-plots-python
def cumulative_boxplot():
    data = adata_copy[adata_copy.obs.age == "A",].obs.n_counts
    # evaluate the histogram
    values, base = np.histogram(data, bins=10)
    # evaluate the cumulative
    cumulative = np.cumsum(values)
    # plot the cumulative function
    plt.scatter(base[:-1], cumulative / cumulative[-1], c="blue")
    # plot the survival function
    # plt.plot(base[:-1], len(data)-cumulative, c='green')



"""
A simple Flyte example.
"""

import typing

from flytekit import task, workflow


@task
def say_hello(name: str = "testing say_hello") -> str:
    """
    A simple Flyte task to say "hello".

    The @task decorator allows Flyte to use this function as a Flyte task, which
    is executed as an isolated, containerized unit of compute.
    """
    return f"hello {name}!"


@task
def greeting_length(greeting: str = "ninechars") -> int:
    """
    A task the counts the length of a greeting.
    """
    return len(greeting)


@workflow
def wf(name: str = "union") -> typing.Tuple[str, int]:
    """
    Declare workflow called `wf`.

    The @workflow decorator defines an execution graph that is composed of tasks
    and potentially sub-workflows. In this simple example, the workflow is
    composed of just one task.

    There are a few important things to note about workflows:
    - Workflows are a domain-specific language (DSL) for creating execution
      graphs and therefore only support a subset of Python's behavior.
    - Tasks must be invoked with keyword arguments
    - The output variables of tasks are Promises, which are placeholders for
      values that are yet to be materialized, not the actual values.
    """
    greeting = say_hello(name=name)
    greeting_len = greeting_length(greeting=greeting)
    return greeting, greeting_len


if __name__ == "__main__":
    # Execute the workflow, simply by invoking it like a function and passing in
    # the necessary parameters
    print(f"Running wf() { wf(name='passengers') }")



import math
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import mlflow
import numpy as np
import pyro
import scipy
import torch
from pyro.infer import Trace_ELBO
from pyro.infer import TraceEnum_ELBO
from pyro.infer.autoguide.guides import AutoGuideList
from pyro.optim.clipped_adam import ClippedAdam
from pyro.optim.optim import PyroOptim
from scvi.dataloaders import DataSplitter
from scvi.train import PyroTrainingPlan
from scvi.train import TrainRunner

from pyrovelocity._velocity_module import VelocityModule


class VelocityAdam(ClippedAdam):
    def step(self, closure: Optional[Callable] = None) -> Optional[Any]:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            group["lr"] *= group["lrd"]

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                grad.clamp_(-group["clip_norm"], group["clip_norm"])
                # freeze non-velocity gene gradient
                grad[grad.isnan()] = 0.0
                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(grad)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(grad)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                if group["weight_decay"] != 0:
                    grad = grad.add(p.data, alpha=group["weight_decay"])

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                denom = exp_avg_sq.sqrt().add_(group["eps"])

                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]
                step_size = (
                    group["lr"] * math.sqrt(bias_correction2) / bias_correction1
                )

                p.data.addcdiv_(exp_avg, denom, value=-step_size)
        return loss


def VelocityClippedAdam(optim_args: Dict[str, float]) -> PyroOptim:
    """
    Wraps :class:`pyro.optim.clipped_adam.ClippedAdam` with :class:`~pyro.optim.optim.PyroOptim`.
    """
    return PyroOptim(VelocityAdam, optim_args)


class EnumTrainingPlan(PyroTrainingPlan):
    def __init__(
        self,
        pyro_velocity: VelocityModule,
        optim: Optional[pyro.optim.PyroOptim] = None,
    ):
        super().__init__(
            pyro_velocity,
            TraceEnum_ELBO(strict_enumeration_warning=True),
            optim,
        )
        self.svi = pyro.infer.SVI(
            model=self.module.model,
            guide=self.module.guide,
            optim=self.optim,
            loss=self.loss_fn,
        )
        self.n_elem = self.module.num_genes * self.module.num_cells * 2

    def training_step(self, batch, batch_idx, optimizer_idx=0):
        args, kwargs = self.module._get_fn_args_from_batch(batch)
        loss = self.svi.step(*args, **kwargs)
        return {
            "train_step_loss": loss,
            "num_elem": args[0].shape[0] * args[0].shape[1] * 2,
        }

    def training_epoch_end(self, outputs):
        n_batch, elbo = 0, 0
        n_elem = 0

        for tensors in outputs:
            elbo += tensors["train_step_loss"]
            n_batch += 1
            # n_elem += tensors['num_elem']
        self.log("elbo_train", elbo / n_batch, prog_bar=True, on_epoch=True)
        # self.log("elbo_train", elbo / self.n_elem, prog_bar=True, on_epoch=True)
        # self.log("elbo_train", elbo / n_elem, prog_bar=True, on_epoch=True)

    def validation_step(self, batch, batch_idx):
        args, kwargs = self.pyro_velocity._get_fn_args_from_batch(batch)
        loss = self.svi.evaluate_loss(*args, **kwargs)
        return {
            "valid_step_loss": loss,
            "num_elem": args[0].shape[0] * args[0].shape[1] * 2,
        }

    def validation_epoch_end(self, outputs):
        """Aggregate validation step information."""
        n_batch, elbo = 0, 0
        n_elem = 0
        for tensors in outputs:
            elbo += tensors["valid_step_loss"]
            n_batch += 1
            ##n_elem += tensors['num_elem']
        self.log(
            "elbo_validation", elbo / n_batch, prog_bar=True, on_epoch=True
        )
        # self.log("elbo_validation", elbo / self.n_elem, prog_bar=True, on_epoch=True)
        ##self.log("elbo_validation", elbo / n_elem, prog_bar=True, on_epoch=True)


class VelocityTrainingMixin:
    def train(
        self,
        use_gpu: Optional[Union[str, int, bool]] = 0,
        early_stopping: bool = False,
        seed: int = 99,
        lr: float = 1e-3,
        train_size: float = 1.0,
        valid_size: float = 0.0,
        batch_size: int = 256,
        max_epochs: int = 100,
        check_val_every_n_epoch: Optional[int] = 1,
        patience: int = 10,
        min_delta: float = 0.0,
        **kwargs,
    ):
        print("base train function")
        pyro.clear_param_store()
        pyro.set_rng_seed(seed)

        data_splitter = DataSplitter(
            self.adata_manager,
            train_size=train_size,
            validation_size=valid_size,
            batch_size=batch_size,
            use_gpu=use_gpu,
        )
        training_plan = EnumTrainingPlan(
            self.module, VelocityClippedAdam({"lr": lr, "lrd": 0.9999})
        )
        runner = TrainRunner(
            self,
            training_plan=training_plan,
            data_splitter=data_splitter,
            check_val_every_n_epoch=check_val_every_n_epoch,
            max_epochs=max_epochs,
            use_gpu=use_gpu,
            early_stopping=early_stopping,
            early_stopping_patience=patience,
            early_stopping_min_delta=min_delta,
            **kwargs,
        )
        return runner()

    def train_faster(
        self,
        use_gpu: Optional[Union[str, int, bool]] = 0,
        seed: int = 99,
        lr: float = 0.01,
        max_epochs: int = 5000,
        log_every: int = 100,
        patient_init: int = 45,
        patient_improve: float = 0.001,
    ) -> List[float]:
        """this method input all adata input gpu for faster IO,
        increase larger dataset by 5-6 fold, however should not
        be used for >20k cells with less than 40GB GPU memory,
        this ignores validation cells"""
        print("train_faster")
        if (use_gpu is False) or (use_gpu == "cpu") or (use_gpu == -1):
            device = "cpu"
        else:
            device = f"cuda:{use_gpu}"
        pyro.clear_param_store()
        pyro.set_rng_seed(seed)
        pyro.enable_validation(True)
        optim = VelocityClippedAdam({"lr": lr, "lrd": 0.1 ** (1 / max_epochs)})
        self.module._model = self.module._model.to(device)
        print("TraceEnum")
        svi = pyro.infer.SVI(
            self.module._model,
            self.module._guide,
            optim,
            # TraceEnum_ELBO(strict_enumeration_warning=True),
            Trace_ELBO(strict_enumeration_warning=True),
        )

        normalizer = self.adata.shape[0] * self.adata.shape[1] * 2
        u = torch.tensor(
            np.array(
                self.adata.layers["raw_unspliced"].toarray(), dtype="float32"
            )
            if scipy.sparse.issparse(self.adata.layers["raw_unspliced"])
            else self.adata.layers["raw_unspliced"],
            dtype=torch.float32,
        ).to(device)
        s = torch.tensor(
            np.array(
                self.adata.layers["raw_spliced"].toarray(), dtype="float32"
            )
            if scipy.sparse.issparse(self.adata.layers["raw_spliced"])
            else self.adata.layers["raw_spliced"],
            dtype=torch.float32,
        ).to(device)

        epsilon = 1e-6

        u_library = torch.tensor(
            # np.array(self.adata.obs.u_lib_size, dtype="float32"), dtype=torch.float32
            np.array(
                np.log(self.adata.obs.u_lib_size_raw + epsilon), dtype="float32"
            ),
            dtype=torch.float32,
        ).to(device)
        s_library = torch.tensor(
            # np.array(self.adata.obs.s_lib_size, dtype="float32"), dtype=torch.float32
            np.array(
                np.log(self.adata.obs.s_lib_size_raw + epsilon), dtype="float32"
            ),
            dtype=torch.float32,
        ).to(device)
        u_library_mean = (
            torch.tensor(
                # np.array(self.adata.obs.u_lib_size_mean, dtype="float32"),
                np.mean(np.log(self.adata.obs.u_lib_size_raw + epsilon)),
                dtype=torch.float32,
            )
            .expand(u_library.shape)
            .to(device)
        )
        s_library_mean = (
            torch.tensor(
                # np.array(self.adata.obs.s_lib_size_mean, dtype="float32"),
                np.mean(np.log(self.adata.obs.s_lib_size_raw + epsilon)),
                dtype=torch.float32,
            )
            .expand(u_library.shape)
            .to(device)
        )
        u_library_scale = (
            torch.tensor(
                # np.array(self.adata.obs.u_lib_size_scale, dtype="float32"),
                np.std(np.log(self.adata.obs.u_lib_size_raw + epsilon)),
                dtype=torch.float32,
            )
            .expand(u_library.shape)
            .to(device)
        )
        s_library_scale = (
            torch.tensor(
                # np.array(self.adata.obs.s_lib_size_scale, dtype="float32"),
                np.std(np.log(self.adata.obs.s_lib_size_raw + epsilon)),
                dtype=torch.float32,
            )
            .expand(u_library.shape)
            .to(device)
        )

        print(u_library_scale.shape)
        print(u.shape)
        print(u_library.shape)
        if "pyro_cell_state" in self.adata.obs.columns:
            cell_state = torch.tensor(
                np.array(self.adata.obs.pyro_cell_state, dtype="float32"),
                dtype=torch.float32,
            ).to(device)
        else:
            cell_state = None

        losses = []
        patience = patient_init
        for step in range(max_epochs):
            if cell_state is None:
                elbos = (
                    svi.step(
                        u,
                        s,
                        u_library.reshape(-1, 1),
                        s_library.reshape(-1, 1),
                        u_library_mean.reshape(-1, 1),
                        s_library_mean.reshape(-1, 1),
                        u_library_scale.reshape(-1, 1),
                        s_library_scale.reshape(-1, 1),
                        None,
                        None,
                    )
                    / normalizer
                )
            else:
                elbos = (
                    svi.step(
                        u,
                        s,
                        u_library.reshape(-1, 1),
                        s_library.reshape(-1, 1),
                        u_library_mean.reshape(-1, 1),
                        s_library_mean.reshape(-1, 1),
                        u_library_scale.reshape(-1, 1),
                        s_library_scale.reshape(-1, 1),
                        None,
                        cell_state.reshape(-1, 1),
                    )
                    / normalizer
                )
            if (step == 0) or (
                ((step + 1) % log_every == 0) and ((step + 1) < max_epochs)
            ):
                mlflow.log_metric("-ELBO", -elbos, step=step + 1)
                print(
                    f"step {step + 1: >4d} loss = {elbos:0.6g} patience = {patience}"
                )
            if step > log_every:
                if (losses[-1] - elbos) < losses[-1] * patient_improve:
                    patience -= 1
                else:
                    patience = patient_init
            if patience <= 0:
                break
            losses.append(elbos)
        mlflow.log_metric("-ELBO", -elbos, step=step + 1)
        mlflow.log_metric("real_epochs", step + 1)
        print(f"step {step + 1: >4d} loss = {elbos:0.6g} patience = {patience}")
        return losses

    def train_faster_with_batch(
        self,
        use_gpu: Optional[Union[str, int, bool]] = 0,
        seed: int = 99,
        lr: float = 1e-2,
        max_epochs: int = 5000,
        log_every: int = 100,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = None,
        new_valid_guide: Optional[AutoGuideList] = None,
        patient_init: int = 45,
        patient_improve: float = 0.0,
        elbo_name: str = "-ELBO",
    ):
        print("train_faster_with_batch")
        if (use_gpu is False) or (use_gpu == "cpu") or (use_gpu == -1):
            device = "cpu"
        else:
            device = f"cuda:{use_gpu}"
        pyro.clear_param_store()
        pyro.set_rng_seed(seed)
        pyro.enable_validation(True)

        adata = self._validate_anndata(self.adata)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size
        )
        optim = VelocityClippedAdam({"lr": lr, "lrd": 0.1 ** (1 / max_epochs)})
        self.module._model = self.module._model.to(device)

        if new_valid_guide is None:
            svi = pyro.infer.SVI(
                self.module._model,
                self.module._guide,
                optim,
                Trace_ELBO(strict_enumeration_warning=True),
            )
        else:
            svi = pyro.infer.SVI(
                self.module._model,
                new_valid_guide,
                optim,
                Trace_ELBO(strict_enumeration_warning=True),
            )

        losses = []
        patience = patient_init
        # normalizer = self.adata.shape[0] * self.adata.shape[1] * 2
        for step in range(max_epochs):
            n_batch = 0
            elbos = 0
            for tensor_dict in scdl:
                args, kwargs = self.module._get_fn_args_from_batch(tensor_dict)
                args = [a.to(device) if a is not None else a for a in args]
                loss = svi.step(*args, **kwargs)
                elbos += loss
                n_batch += 1
            # elbos = elbos / normalizer #n_batch
            elbos = elbos / n_batch
            if (step == 0) or (
                ((step + 1) % log_every == 0) and ((step + 1) < max_epochs)
            ):
                mlflow.log_metric("-ELBO", -elbos, step=step + 1)
                print(
                    f"step {step + 1: >4d} loss = {elbos:0.6g} patience = {patience}"
                )
            if step > log_every:
                if (losses[-1] - elbos) < losses[-1] * patient_improve:
                    patience -= 1
                    # print(f"step {step: >4d} loss = {elbos:0.6g} patience = {patience}")
                else:
                    patience = patient_init
            if patience <= 0:
                break
            losses.append(elbos)
        mlflow.log_metric("-ELBO", -elbos, step=step + 1)
        mlflow.log_metric("real_epochs", step + 1)
        print(f"step {step: >4d} loss = {elbos:0.6g} patience = {patience}")
        return losses



from typing import Tuple

import anndata
import numpy as np
import pandas as pd


def calculate_criticality_index(
    adata: anndata.AnnData,
) -> Tuple[float, float, float, float]:
    """
    Calculates a criticality index for multimodal single-cell RNA-seq
    data stored in an AnnData object.

    Parameters:
        adata (anndata.AnnData): The AnnData object containing single-cell
                                 RNA-seq data with unspliced and spliced layers.

    Returns:
        criticality_index (float): The criticality index.
        pcc_mean (float): Mean of the Pearson correlation coefficient.
        sd_mean (float): Mean of the standard deviation.
        H_mean (float): Mean of the conditional entropy.

    Examples:
        >>> import scvelo as scv
        >>> adata = scv.datasets.simulation(
        ...     random_seed=0,
        ...     n_obs=10,
        ...     n_vars=3,
        ...     alpha=5,
        ...     beta=0.5,
        ...     gamma=0.3,
        ...     alpha_=0,
        ...     switches=[1, 5, 10],
        ...     noise_model="gillespie",
        ... )
        >>> print(f"unspliced\\n{adata.layers['unspliced'].T}")
        >>> print(f"spliced\\n{adata.layers['spliced'].T}")
        >>> (
        ...     criticality_index,
        ...     pcc_mean,
        ...     sd_mean,
        ...     H_mean
        ... ) = calculate_criticality_index(adata)
        >>> print(f"PCC Mean: {pcc_mean}")
        >>> print(f"SD Mean: {sd_mean}")
        >>> print(f"H Mean: {H_mean}")
        >>> print(f"Criticality Index: {criticality_index:.4f}")
    """

    if adata.X.size == 0:
        raise ValueError("AnnData object must not be empty.")

    if "unspliced" not in adata.layers or "spliced" not in adata.layers:
        raise KeyError(
            "AnnData object must contain both 'unspliced' and 'spliced' layers."
        )

    unspliced_data = adata.layers["unspliced"].T
    spliced_data = adata.layers["spliced"].T

    pcc_unspliced = np.corrcoef(unspliced_data)
    pcc_spliced = np.corrcoef(spliced_data)

    pcc_mean = np.mean(
        (
            pcc_unspliced[np.triu_indices(unspliced_data.shape[0], k=1)]
            + pcc_spliced[np.triu_indices(spliced_data.shape[0], k=1)]
        )
        / 2
    )

    sd_mean = np.mean(
        np.hstack(
            (np.std(unspliced_data, axis=1), np.std(spliced_data, axis=1))
        )
    )

    def conditional_entropy(data):
        df = pd.DataFrame(data)
        joint_prob = (
            df.apply(lambda x: x.value_counts() / len(x))
            .fillna(0)
            .values.flatten()
        )
        joint_prob = joint_prob.reshape(data.shape[1], -1)
        marginal_prob = np.sum(joint_prob, axis=1)
        cond_prob = joint_prob / marginal_prob[:, None]
        cond_entropy = -np.nansum(joint_prob * np.log2(cond_prob))
        return cond_entropy

    H_unspliced = conditional_entropy(unspliced_data)
    H_spliced = conditional_entropy(spliced_data)

    H_mean = (H_unspliced + H_spliced) / 2

    criticality_index = (pcc_mean * sd_mean) / H_mean
    return criticality_index, pcc_mean, sd_mean, H_mean



import argparse
from pathlib import Path
from typing import Sequence, Union

import hydra
import rich.console
import rich.pretty
import rich.syntax
import rich.tree
from hydra.core.config_store import ConfigStore
from hydra.core.global_hydra import GlobalHydra
from hydra_zen import make_config, make_custom_builds_fn
from omegaconf import DictConfig, ListConfig, OmegaConf
from pytorch_lightning.utilities import rank_zero_only

from pyrovelocity.api import train_model
from pyrovelocity.utils import get_pylogger


def hydra_zen_configure():
    """
    Use hydra-zen to generate a configuration for hydra's config store.
    Generates a hydra-zen configuration for the pyrovelocity pipeline,
    which includes dataset, model training, and reporting configurations.

    This function defines helper functions for creating dataset, model,
    and reports configurations, and uses these functions to generate the
    hydra-zen configuration.

    Returns:
        Config : Type[DataClass]
        See the documentation for hydra_zen.make_config for more information.
    """
    # define helper functions
    pbuilds = make_custom_builds_fn(
        zen_partial=True, populate_full_signature=True
    )

    def create_dataset_config(
        name, dl_root, data_file, rel_path, url, process_method, process_args
    ):
        return dict(
            data_file=data_file,
            dl_root=dl_root,
            dl_path="${.dl_root}/${.data_file}",
            rel_path=rel_path,
            url=url,
            derived=dict(
                process_method=process_method,
                process_args=process_args,
                rel_path="${data_external.processed_path}/"
                + f"{name}_processed.h5ad",
            ),
        )

    def create_model_config(
        source,
        name,
        model_suffix,
        vector_field_basis,
        **custom_training_parameters,
    ):
        return dict(
            path="${paths.models}/" + f"{name}_model{model_suffix}",
            model_path="${.path}/model",
            input_data_path="${data_external."
            + f"{source}.{name}"
            + ".derived.rel_path}",
            trained_data_path="${.path}/trained.h5ad",
            pyrovelocity_data_path="${.path}/pyrovelocity.pkl",
            metrics_path="${.path}/metrics.json",
            run_info_path="${.path}/run_info.json",
            vector_field_parameters=dict(basis=vector_field_basis),
            training_parameters=pbuilds(
                train_model,
                loss_plot_path="${..path}/loss_plot.png",
                patient_improve=0.0001,
                patient_init=45,
                **custom_training_parameters,
            ),
        )

    def create_reports_config(model_name: str, model_number: int):
        path = f"{paths['reports']}/{model_name}_model{model_number}"
        return dict(
            path=path,
            dataframe_path=f"{paths['data']}/processed/{model_name}_model{model_number}_dataframe.pkl.zst",
            shared_time_plot=f"{path}/shared_time.pdf",
            volcano_plot=f"{path}/volcano.pdf",
            rainbow_plot=f"{path}/rainbow.pdf",
            uncertainty_param_plot=f"{path}/param_uncertainties.pdf",
            vector_field_plot=f"{path}/vector_field.pdf",
            biomarker_selection_plot=f"{path}/markers_selection_scatterplot.tif",
            biomarker_phaseportrait_plot=f"{path}/markers_phaseportrait.pdf",
        )

    # define configuration
    base = dict(log_level="INFO")

    paths = dict(
        data="data",
        models="models",
        reports="reports",
    )

    config = make_config(
        base=base,
        paths=paths,
        data_external=dict(
            root_path="${paths.data}/external",
            processed_path="${paths.data}/processed",
            sources=["simulate", "velocyto", "scvelo", "pyrovelocity"],
            simulate=dict(
                download=["medium"],
                process=["medium"],
                sources=dict(
                    gcs_root_url="https://storage.googleapis.com/pyrovelocity/data"
                ),
                medium=create_dataset_config(
                    "simulated_medium",
                    dl_root="${data_external.root_path}",
                    data_file="simulated_medium.h5ad",
                    rel_path="${data_external.root_path}/simulated_medium.h5ad",
                    url="${data_external.simulate.sources.gcs_root_url}/simulated_medium.h5ad",
                    process_method="load_data",
                    process_args=dict(),
                ),
            ),
            velocyto=dict(
                download=["pons"],
                process=["pons"],
                sources=dict(
                    gcs_root_url="https://storage.googleapis.com/pyrovelocity/data"
                ),
                pons=create_dataset_config(
                    "pons",
                    dl_root="${data_external.root_path}",
                    data_file="oligo_lite.h5ad",
                    rel_path="${data_external.root_path}/oligo_lite.h5ad",
                    url="${data_external.velocyto.sources.gcs_root_url}/oligo_lite.h5ad",
                    process_method="load_data",
                    process_args=dict(),
                ),
            ),
            scvelo=dict(
                download=["pancreas", "pbmc68k"],
                process=["pancreas", "pbmc68k"],
                sources=dict(
                    figshare_root_url="https://ndownloader.figshare.com/files",
                    scvelo_root_url="https://github.com/theislab/scvelo_notebooks/raw/master",
                ),
                pancreas=create_dataset_config(
                    "pancreas",
                    dl_root="data/Pancreas",
                    data_file="endocrinogenesis_day15.h5ad",
                    rel_path="${data_external.root_path}/endocrinogenesis_day15.h5ad",
                    url="${data_external.scvelo.sources.scvelo_root_url}/data/Pancreas/endocrinogenesis_day15.h5ad",
                    process_method="load_data",
                    process_args=dict(process_cytotrace=True),
                ),
                pbmc68k=create_dataset_config(
                    "pbmc68k",
                    dl_root="data/PBMC",
                    data_file="pbmc68k.h5ad",
                    rel_path="${data_external.root_path}/pbmc68k.h5ad",
                    url="${data_external.scvelo.sources.figshare_root_url}/27686886",
                    process_method="load_data",
                    process_args=dict(),
                ),
            ),
            pyrovelocity=dict(
                download=["larry", "pbmc10k"],
                process=["pbmc10k"],
                sources=dict(
                    figshare_root_url="https://ndownloader.figshare.com/files"
                ),
                larry=create_dataset_config(
                    "larry",
                    dl_root="${data_external.root_path}",
                    data_file="larry.h5ad",
                    rel_path="${data_external.root_path}/larry.h5ad",
                    url="${data_external.pyrovelocity.sources.figshare_root_url}/37028569",
                    process_method="load_data",
                    process_args=dict(),
                ),
                larry_tips=create_dataset_config(
                    "larry_tips",
                    dl_root="${data_external.root_path}",
                    data_file="larry_tips.h5ad",
                    rel_path="${data_external.root_path}/larry_tips.h5ad",
                    url="${data_external.pyrovelocity.sources.figshare_root_url}/37028569",
                    process_method="load_data",
                    process_args=dict(),
                ),
                larry_mono=create_dataset_config(
                    "larry_mono",
                    dl_root="${data_external.root_path}",
                    data_file="larry_mono.h5ad",
                    rel_path="${data_external.root_path}/larry_mono.h5ad",
                    url="${data_external.pyrovelocity.sources.figshare_root_url}/37028572",
                    process_method="load_data",
                    process_args=dict(),
                ),
                larry_neu=create_dataset_config(
                    "larry_neu",
                    dl_root="${data_external.root_path}",
                    data_file="larry_neu.h5ad",
                    rel_path="${data_external.root_path}/larry_neu.h5ad",
                    url="${data_external.pyrovelocity.sources.figshare_root_url}/37028575",
                    process_method="load_data",
                    process_args=dict(),
                ),
                larry_multilineage=create_dataset_config(
                    "larry_multilineage",
                    dl_root="${data_external.root_path}",
                    data_file="larry_mono.h5ad",
                    rel_path="${data_external.root_path}/larry_mono.h5ad",
                    url="${data_external.pyrovelocity.sources.figshare_root_url}/37028572",
                    process_method="load_data",
                    process_args=dict(),
                ),
                pbmc10k=create_dataset_config(
                    "pbmc10k",
                    dl_root="${data_external.root_path}",
                    data_file="pbmc10k.h5ad",
                    rel_path="${data_external.root_path}/pbmc10k.h5ad",
                    url="${data_external.pyrovelocity.sources.figshare_root_url}/pbmc10k",
                    process_method="load_data",
                    process_args=dict(),
                ),
            ),
        ),
        model_training=dict(
            train=[
                "simulate_model1",
                "simulate_model2",
                "pancreas_model1",
                "pancreas_model2",
                "pbmc68k_model1",
                "pbmc68k_model2",
                "pons_model1",
                "pons_model2",
                "larry_model2",
                "larry_tips_model2",
                "larry_mono_model2",
                "larry_neu_model2",
                "larry_multilineage_model2",
                "pbmc10k_model2",
            ],
            simulate_model1=create_model_config(
                "simulate",
                "medium",
                1,
                "umap",
                cell_state="leiden",
                guide_type="auto_t0_constraint",
                max_epochs=4000,
            ),
            simulate_model2=create_model_config(
                "simulate",
                "medium",
                2,
                "umap",
                cell_state="leiden",
                max_epochs=4000,
                offset=True,
            ),
            pancreas_model1=create_model_config(
                "scvelo",
                "pancreas",
                1,
                "umap",
                guide_type="auto_t0_constraint",
                max_epochs=2000,
            ),
            pancreas_model2=create_model_config(
                "scvelo",
                "pancreas",
                2,
                "umap",
                offset=True,
                max_epochs=2000,
            ),
            pbmc68k_model1=create_model_config(
                "scvelo",
                "pbmc68k",
                1,
                "tsne",
                guide_type="auto_t0_constraint",
                cell_state="celltype",
                max_epochs=2000,
            ),
            pbmc68k_model2=create_model_config(
                "scvelo",
                "pbmc68k",
                2,
                "tsne",
                cell_state="celltype",
                offset=True,
                max_epochs=2000,
            ),
            pons_model1=create_model_config(
                "velocyto",
                "pons",
                1,
                "umap",
                guide_type="auto_t0_constraint",
                cell_state="celltype",
                max_epochs=2000,
            ),
            pons_model2=create_model_config(
                "velocyto",
                "pons",
                2,
                "umap",
                cell_state="celltype",
                offset=True,
                max_epochs=2000,
            ),
            larry_model2=create_model_config(
                "pyrovelocity",
                "larry",
                2,
                "emb",
                svi_train=True,
                batch_size=4000,
                cell_state="state_info",
                offset=True,
                max_epochs=1000,
            ),
            larry_tips_model2=create_model_config(
                "pyrovelocity",
                "larry_tips",
                2,
                "umap",
                svi_train=True,
                batch_size=4000,
                cell_state="state_info",
                offset=True,
                max_epochs=1000,
            ),
            larry_mono_model2=create_model_config(
                "pyrovelocity",
                "larry_mono",
                2,
                "emb",
                svi_train=True,
                batch_size=4000,
                cell_state="state_info",
                offset=True,
                max_epochs=1000,
            ),
            larry_neu_model2=create_model_config(
                "pyrovelocity",
                "larry_neu",
                2,
                "emb",
                svi_train=True,
                batch_size=4000,
                cell_state="state_info",
                offset=True,
                max_epochs=1000,
            ),
            larry_multilineage_model2=create_model_config(
                "pyrovelocity",
                "larry_multilineage",
                2,
                "emb",
                svi_train=True,
                batch_size=4000,
                cell_state="state_info",
                offset=True,
                max_epochs=1000,
            ),
            pbmc10k_model2=create_model_config(
                "pyrovelocity",
                "pbmc10k",
                2,
                "umap",
                cell_state="celltype",
                offset=True,
                max_epochs=2000,
            ),
        ),
        reports=dict(
            model_summary=dict(
                summarize=[
                    "simulate_model1",
                    "simulate_model2",
                    "pancreas_model1",
                    "pancreas_model2",
                    "pbmc68k_model1",
                    "pbmc68k_model2",
                    "pons_model1",
                    "pons_model2",
                    "pbmc10k_model2",
                    "larry_tips_model2",
                ],
                simulate_model1=create_reports_config("medium", 1),
                simulate_model2=create_reports_config("medium", 2),
                pancreas_model1=create_reports_config("pancreas", 1),
                pancreas_model2=create_reports_config("pancreas", 2),
                pbmc68k_model1=create_reports_config("pbmc68k", 1),
                pbmc68k_model2=create_reports_config("pbmc68k", 2),
                pons_model1=create_reports_config("pons", 1),
                pons_model2=create_reports_config("pons", 2),
                pbmc10k_model2=create_reports_config("pbmc10k", 2),
                larry_tips_model2=create_reports_config("larry_tips", 2),
            ),
            figure2=dict(
                tag="fig2",
                path="${paths.reports}/${.tag}",
                tif_path="${.path}/${.tag}_raw_gene_selection_model1.tif",
                svg_path="${.path}/${.tag}_raw_gene_selection_model1.svg",
                biomarker_selection_path="${.path}/${.tag}_markers_selection_scatterplot.tif",
                biomarker_phaseportrait_path="${.path}/${.tag}_markers_phaseportrait.pdf",
                uncertainty_param_plot_path="${.path}/${.tag}_param_uncertainties.pdf",
                uncertainty_magnitude_plot_path="${.path}/${.tag}_magnitude_uncertainties.pdf",
                pancreas_model1=dict(
                    shared_time_plot="${..path}/${..tag}_pancreas_shared_time.pdf",
                    volcano_plot="${..path}/${..tag}_pancreas_volcano.pdf",
                    rainbow_plot="${..path}/${..tag}_pancreas_rainbow.pdf",
                    vector_field_plot="${..path}/${..tag}_pancreas_vector_field.pdf",
                ),
            ),
            figure2_extras=dict(
                tag="fig2",
                path="${paths.reports}/${.tag}",
                shared_time_plot="${.path}/figure2_extras.pdf",
            ),
            figureS3=dict(
                tag="figS3",
                path="${paths.reports}/${.tag}",
                tif_path="${.path}/${.tag}_raw_gene_selection_model2.tif",
                svg_path="${.path}/${.tag}_raw_gene_selection_model2.svg",
            ),
            figureS3_extras=dict(
                tag="figS3",
                path="${paths.reports}/${.tag}",
                violin_plots_pbmc_lin="${.path}/figureS3_extras_pbmc_lin.pdf",
                violin_plots_pbmc_log="${.path}/figureS3_extras_pbmc_log.pdf",
                violin_plots_larry_lin="${.path}/figureS3_extras_larry_lin.pdf",
                violin_plots_larry_log="${.path}/figureS3_extras_larry_log.pdf",
            ),
        ),
    )
    return config


def hydra_zen_compressed_configure():
    """
    Use hydra-zen to generate a configuration for hydra's config store.
    Generates a hydra-zen configuration for the pyrovelocity pipeline,
    which includes dataset, model training, and reporting configurations.

    This function defines helper functions for creating dataset, model,
    and reports configurations, and uses these functions to generate the
    hydra-zen configuration.

    Returns:
        Config : Type[DataClass]
        See the documentation for hydra_zen.make_config for more information.
    """
    # define helper functions
    pbuilds = make_custom_builds_fn(
        zen_partial=True, populate_full_signature=True
    )

    def create_dataset_config(
        source,
        name,
        dl_root,
        data_file,
        rel_path,
        url,
        process_method,
        process_args,
    ):
        return dict(
            source=source,
            data_file=data_file,
            dl_root=dl_root,
            dl_path="${.dl_root}/${.data_file}",
            rel_path=rel_path,
            url=url,
            derived=dict(
                process_method=process_method,
                process_args=process_args,
                rel_path="${paths.data_processed}/" + f"{name}_processed.h5ad",
                thresh_histogram_path="${paths.data_processed}/"
                + f"{name}_thresh_histogram.pdf",
            ),
        )

    def create_model_config(
        source,
        name,
        model_suffix,
        vector_field_basis,
        gpu_id=0,
        **custom_training_parameters,
    ):
        return dict(
            path="${paths.models}/" + f"{name}_model{model_suffix}",
            model_path="${.path}/model",
            input_data_path="${data_sets." + f"{name}" + ".derived.rel_path}",
            trained_data_path="${.path}/trained.h5ad",
            pyrovelocity_data_path="${.path}/pyrovelocity.pkl.zst",
            posterior_samples_path="${.path}/posterior_samples.pkl.zst",
            metrics_path="${.path}/metrics.json",
            run_info_path="${.path}/run_info.json",
            vector_field_parameters=dict(basis=vector_field_basis),
            gpu_id=gpu_id,
            training_parameters=pbuilds(
                train_model,
                loss_plot_path="${..path}/loss_plot.png",
                patient_improve=0.0001,
                patient_init=45,
                seed="${base.seed}",
                **custom_training_parameters,
            ),
        )

    def create_reports_config(model_name: str, model_number: int):
        path = f"{paths['reports']}/{model_name}_model{model_number}"
        return dict(
            path=path,
            trained_data_path="${paths.models}/"
            + f"{model_name}_model{model_number}/trained.h5ad",
            pyrovelocity_data_path="${paths.models}/"
            + f"{model_name}_model{model_number}/pyrovelocity.pkl.zst",
            dataframe_path=f"{paths['data']}/processed/{model_name}_model{model_number}_dataframe.pkl.zst",
            shared_time_plot=f"{path}/shared_time.pdf",
            volcano_plot=f"{path}/volcano.pdf",
            rainbow_plot=f"{path}/rainbow.pdf",
            uncertainty_param_plot=f"{path}/param_uncertainties.pdf",
            vector_field_plot=f"{path}/vector_field.pdf",
            posterior_phase_portraits=f"{path}/posterior_phase_portraits",
            t0_selection=f"{path}/t0_selection.tif",
            biomarker_selection_plot=f"{path}/markers_selection_scatterplot.tif",
            biomarker_phaseportrait_plot=f"{path}/markers_phaseportrait.pdf",
            fig2_part1_plot=f"{path}/fig2_part1_plot.pdf",
            fig2_part2_plot=f"{path}/fig2_part2_plot.pdf",
            violin_clusters_lin=f"{path}/clusters_violin_lin.pdf",
            violin_clusters_log=f"{path}/clusters_violin_log.pdf",
        )

    base = dict(
        log_level="INFO",
        count_threshold=0,
        seed=99,
    )

    paths = dict(
        data="data",
        models="models",
        reports="reports",
        data_external="${paths.data}/external",
        data_processed="${paths.data}/processed",
    )

    download_data = [
        "simulate_medium",
        "pons",
        "pancreas",
        "bonemarrow",
        "pbmc5k",
        "pbmc10k",
        "pbmc68k",
        "larry",
        "larry_cospar",
        "larry_cytotrace",
        "larry_dynamical",
    ]

    process_data = [
        "simulate_medium",
        "pons",
        "pancreas",
        "bonemarrow",
        "pbmc5k",
        "pbmc10k",
        "pbmc68k",
        "larry",
        "larry_tips",
        "larry_mono",
        "larry_neu",
        "larry_multilineage",
    ]
    train_models = [
        "pancreas_model2",
        "pbmc68k_model2",
        "pons_model2",
        "larry_model2",
        "larry_tips_model2",
        "larry_mono_model2",
        "larry_neu_model2",
        "larry_multilineage_model2",
        "pbmc10k_model2",
        "pbmc5k_model2",
    ]

    model_training = dict(
        simulate_model1=create_model_config(
            "simulate",
            "simulate_medium",
            1,
            "umap",
            cell_state="leiden",
            guide_type="auto_t0_constraint",
            max_epochs=4000,
        ),
        simulate_model2=create_model_config(
            "simulate",
            "simulate_medium",
            2,
            "umap",
            cell_state="leiden",
            max_epochs=300,
            offset=True,
        ),
        pancreas_model1=create_model_config(
            "scvelo",
            "pancreas",
            1,
            "umap",
            gpu_id=0,
            guide_type="auto_t0_constraint",
            max_epochs=2000,
        ),
        pancreas_model2=create_model_config(
            "scvelo",
            "pancreas",
            2,
            "umap",
            gpu_id=0,
            offset=True,
            max_epochs=2000,
        ),
        bonemarrow_model1=create_model_config(
            "scvelo",
            "bonemarrow",
            1,
            "umap",
            gpu_id=0,
            guide_type="auto_t0_constraint",
            max_epochs=2000,
        ),
        bonemarrow_model2=create_model_config(
            "scvelo",
            "bonemarrow",
            2,
            "umap",
            gpu_id=0,
            offset=True,
            max_epochs=2000,
        ),
        pbmc68k_model1=create_model_config(
            "scvelo",
            "pbmc68k",
            1,
            "tsne",
            gpu_id=1,
            guide_type="auto_t0_constraint",
            cell_state="celltype",
            max_epochs=2000,
        ),
        pbmc68k_model2=create_model_config(
            "scvelo",
            "pbmc68k",
            2,
            "tsne",
            gpu_id=1,
            cell_state="celltype",
            offset=True,
            max_epochs=2000,
        ),
        pons_model1=create_model_config(
            "velocyto",
            "pons",
            1,
            "umap",
            gpu_id=2,
            guide_type="auto_t0_constraint",
            cell_state="celltype",
            max_epochs=2000,
        ),
        pons_model2=create_model_config(
            "velocyto",
            "pons",
            2,
            "umap",
            gpu_id=2,
            cell_state="celltype",
            offset=True,
            max_epochs=2000,
        ),
        larry_model1=create_model_config(
            "pyrovelocity",
            "larry",
            1,
            "emb",
            gpu_id=3,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            guide_type="auto_t0_constraint",
            max_epochs=1000,
        ),
        larry_model2=create_model_config(
            "pyrovelocity",
            "larry",
            2,
            "emb",
            gpu_id=3,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            offset=True,
            max_epochs=1000,
        ),
        larry_tips_model1=create_model_config(
            "pyrovelocity",
            "larry_tips",
            1,
            "umap",
            gpu_id=0,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            guide_type="auto_t0_constraint",
            max_epochs=1000,
        ),
        larry_tips_model2=create_model_config(
            "pyrovelocity",
            "larry_tips",
            2,
            "umap",
            gpu_id=0,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            offset=True,
            max_epochs=1000,
        ),
        larry_mono_model1=create_model_config(
            "pyrovelocity",
            "larry_mono",
            1,
            "emb",
            gpu_id=1,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            guide_type="auto_t0_constraint",
            max_epochs=1000,
        ),
        larry_mono_model2=create_model_config(
            "pyrovelocity",
            "larry_mono",
            2,
            "emb",
            gpu_id=1,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            offset=True,
            max_epochs=1000,
        ),
        larry_neu_model1=create_model_config(
            "pyrovelocity",
            "larry_neu",
            1,
            "emb",
            gpu_id=2,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            guide_type="auto_t0_constraint",
            max_epochs=1000,
        ),
        larry_neu_model2=create_model_config(
            "pyrovelocity",
            "larry_neu",
            2,
            "emb",
            gpu_id=2,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            offset=True,
            max_epochs=1000,
        ),
        larry_multilineage_model1=create_model_config(
            "pyrovelocity",
            "larry_multilineage",
            1,
            "emb",
            gpu_id=3,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            guide_type="auto_t0_constraint",
            max_epochs=1000,
        ),
        larry_multilineage_model2=create_model_config(
            "pyrovelocity",
            "larry_multilineage",
            2,
            "emb",
            gpu_id=3,
            svi_train=True,
            batch_size=4000,
            cell_state="state_info",
            offset=True,
            max_epochs=1000,
        ),
        pbmc10k_model1=create_model_config(
            "pyrovelocity",
            "pbmc10k",
            1,
            "umap",
            gpu_id=1,
            cell_state="celltype",
            guide_type="auto_t0_constraint",
            max_epochs=2000,
        ),
        pbmc10k_model2=create_model_config(
            "pyrovelocity",
            "pbmc10k",
            2,
            "umap",
            gpu_id=1,
            cell_state="celltype",
            offset=True,
            max_epochs=2000,
        ),
        pbmc5k_model1=create_model_config(
            "pyrovelocity",
            "pbmc5k",
            1,
            "umap",
            gpu_id=2,
            cell_state="celltype",
            guide_type="auto_t0_constraint",
            max_epochs=2000,
        ),
        pbmc5k_model2=create_model_config(
            "pyrovelocity",
            "pbmc5k",
            2,
            "umap",
            gpu_id=2,
            cell_state="celltype",
            offset=True,
            max_epochs=2000,
        ),
    )

    data_sets = dict(
        simulate_medium=create_dataset_config(
            source="simulate",
            name="simulated_medium",
            dl_root="${paths.data_external}",
            data_file="simulated_medium.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://storage.googleapis.com/pyrovelocity/data/simulated_medium.h5ad",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        pons=create_dataset_config(
            source="velocyto",
            name="pons",
            dl_root="${paths.data_external}",
            data_file="oligo_lite.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://storage.googleapis.com/pyrovelocity/data/oligo_lite.h5ad",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        bonemarrow=create_dataset_config(
            source="scvelo",
            name="bonemarrow",
            dl_root="data/BoneMarrow",
            data_file="human_cd34_bone_marrow.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://ndownloader.figshare.com/files/27686835",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        pancreas=create_dataset_config(
            source="scvelo",
            name="pancreas",
            dl_root="data/Pancreas",
            data_file="endocrinogenesis_day15.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://github.com/theislab/scvelo_notebooks/raw/master/data/Pancreas/endocrinogenesis_day15.h5ad",
            process_method="load_data",
            process_args=dict(
                process_cytotrace=True, count_thres="${base.count_threshold}"
            ),
        ),
        pbmc68k=create_dataset_config(
            source="scvelo",
            name="pbmc68k",
            dl_root="data/PBMC",
            data_file="pbmc68k.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://ndownloader.figshare.com/files/27686886",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        larry=create_dataset_config(
            source="pyrovelocity",
            name="larry",
            dl_root="${paths.data_external}",
            data_file="larry.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://ndownloader.figshare.com/files/37028569",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        larry_cospar=create_dataset_config(
            source="pyrovelocity",
            name="larry_cospar",
            dl_root="${paths.data_external}",
            data_file="larry_cospar.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://storage.googleapis.com/pyrovelocity/data/larry_cospar.h5ad",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        larry_cytotrace=create_dataset_config(
            source="pyrovelocity",
            name="larry_cytotrace",
            dl_root="${paths.data_external}",
            data_file="larry_cytotrace.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://storage.googleapis.com/pyrovelocity/data/larry_cytotrace.h5ad",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        larry_dynamical=create_dataset_config(
            source="pyrovelocity",
            name="larry_dynamical",
            dl_root="${paths.data_external}",
            data_file="larry_dynamical.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://storage.googleapis.com/pyrovelocity/data/larry_dynamical.h5ad",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        larry_tips=create_dataset_config(
            source="pyrovelocity",
            name="larry_tips",
            dl_root="${paths.data_external}",
            data_file="larry_tips.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://ndownloader.figshare.com/files/37028569",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        larry_mono=create_dataset_config(
            source="pyrovelocity",
            name="larry_mono",
            dl_root="${paths.data_external}",
            data_file="larry_mono.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://ndownloader.figshare.com/files/37028569",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        larry_neu=create_dataset_config(
            source="pyrovelocity",
            name="larry_neu",
            dl_root="${paths.data_external}",
            data_file="larry_neu.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://ndownloader.figshare.com/files/37028575",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        larry_multilineage=create_dataset_config(
            source="pyrovelocity",
            name="larry_multilineage",
            dl_root="${paths.data_external}",
            data_file="larry_multilineage.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://ndownloader.figshare.com/files/37028569",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        pbmc10k=create_dataset_config(
            source="pyrovelocity",
            name="pbmc10k",
            dl_root="${paths.data_external}",
            data_file="pbmc10k.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://storage.googleapis.com/pyrovelocity/data/pbmc10k.h5ad",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
        pbmc5k=create_dataset_config(
            source="pyrovelocity",
            name="pbmc5k",
            dl_root="${paths.data_external}",
            data_file="pbmc5k.h5ad",
            rel_path="${paths.data_external}/${.data_file}",
            url="https://storage.googleapis.com/pyrovelocity/data/pbmc5k.h5ad",
            process_method="load_data",
            process_args=dict(count_thres="${base.count_threshold}"),
        ),
    )

    return make_config(
        base=base,
        paths=paths,
        download_data=download_data,
        process_data=process_data,
        train_models=train_models,
        data_sets=data_sets,
        model_training={
            k: model_training[k]
            for k in model_training
            # if fnmatch.fnmatch(k, "*_model2") or k == "pancreas_model1"
        },
        reports=dict(
            model_summary=dict(
                # simulate_model1=create_reports_config("medium", 1),
                # simulate_model2=create_reports_config("medium", 2),
                pancreas_model1=create_reports_config("pancreas", 1),
                pancreas_model2=create_reports_config("pancreas", 2),
                bonemarrow_model1=create_reports_config("bonemarrow", 1),
                bonemarrow_model2=create_reports_config("bonemarrow", 2),
                pbmc68k_model1=create_reports_config("pbmc68k", 1),
                pbmc68k_model2=create_reports_config("pbmc68k", 2),
                pons_model1=create_reports_config("pons", 1),
                pons_model2=create_reports_config("pons", 2),
                pbmc10k_model1=create_reports_config("pbmc10k", 1),
                pbmc10k_model2=create_reports_config("pbmc10k", 2),
                pbmc5k_model1=create_reports_config("pbmc5k", 1),
                pbmc5k_model2=create_reports_config("pbmc5k", 2),
                larry_tips_model1=create_reports_config("larry_tips", 1),
                larry_tips_model2=create_reports_config("larry_tips", 2),
                larry_mono_model1=create_reports_config("larry_mono", 1),
                larry_mono_model2=create_reports_config("larry_mono", 2),
                larry_neu_model1=create_reports_config("larry_neu", 1),
                larry_neu_model2=create_reports_config("larry_neu", 2),
                larry_multilineage_model1=create_reports_config(
                    "larry_multilineage", 1
                ),
                larry_multilineage_model2=create_reports_config(
                    "larry_multilineage", 2
                ),
                # larry_model1=create_reports_config("larry", 1),
                # larry_model2=create_reports_config("larry", 2),
            ),
            figure2=dict(
                tag="fig2",
                path="${paths.reports}/${.tag}",
                tif_path="${.path}/${.tag}_raw_gene_selection_model1.tif",
                svg_path="${.path}/${.tag}_raw_gene_selection_model1.svg",
                biomarker_selection_path="${.path}/${.tag}_markers_selection_scatterplot.tif",
                biomarker_phaseportrait_path="${.path}/${.tag}_markers_phaseportrait.pdf",
                uncertainty_param_plot_path="${.path}/${.tag}_param_uncertainties.pdf",
                uncertainty_magnitude_plot_path="${.path}/${.tag}_magnitude_uncertainties.pdf",
                pancreas_model1=dict(
                    shared_time_plot="${..path}/${..tag}_pancreas_shared_time.pdf",
                    volcano_plot="${..path}/${..tag}_pancreas_volcano.pdf",
                    rainbow_plot="${..path}/${..tag}_pancreas_rainbow.pdf",
                    vector_field_plot="${..path}/${..tag}_pancreas_vector_field.pdf",
                ),
            ),
            figure2_extras=dict(
                tag="fig2",
                path="${paths.reports}/${.tag}",
                shared_time_plot="${.path}/figure2_extras.pdf",
            ),
            figure3=dict(
                tag="fig3",
                path="${paths.reports}/${.tag}",
                figure3="${.path}/figure3.pdf",
            ),
            figureS2_extra_2=dict(
                tag="figS2",
                path="${paths.reports}/${.tag}",
                mean_length_vs_uncertain="${.path}/${.tag}_mean_length_vs_uncertain.pdf",
            ),
            figureS2=dict(
                tag="figS2",
                path="${paths.reports}/${.tag}",
                rayleigh_classifier_plot="${.path}/${.tag}_rayleigh_classifier.pdf",
                distance_time_correlation_plot="${.path}/${.tag}_distance_time_correlation.pdf",
            ),
            figureS2_extras=dict(
                tag="figS2",
                path="${paths.reports}/${.tag}",
                subset_genes_plot="${.path}/${.tag}_subset3g_pbmc68k.pdf",
                subset_pkl="${.path}/${.tag}_subset3g_pbmc68k.pkl.zst",
            ),
            figureS4=dict(
                tag="figS4",
                path="${paths.reports}/${.tag}",
                figureS4="${.path}/${.tag}.pdf",
            ),
            figureS3=dict(
                tag="figS3",
                path="${paths.reports}/${.tag}",
                tif_path="${.path}/${.tag}_raw_gene_selection_model2.tif",
                svg_path="${.path}/${.tag}_raw_gene_selection_model2.svg",
            ),
            figureS3_extras=dict(
                tag="figS3",
                path="${paths.reports}/${.tag}",
                violin_plots_pbmc_lin="${.path}/figureS3_extras_pbmc_lin.pdf",
                violin_plots_pbmc_log="${.path}/figureS3_extras_pbmc_log.pdf",
                violin_plots_larry_lin="${.path}/figureS3_extras_larry_lin.pdf",
                violin_plots_larry_log="${.path}/figureS3_extras_larry_log.pdf",
            ),
        ),
    )


def test_hydra_zen_configure():
    """
    Use hydra-zen to generate a test configuration for hydra's config store.
    Generates a hydra-zen configuration to test the pyrovelocity pipeline,
    which includes dataset, model training, and reporting configurations.

    This function defines helper functions for creating dataset, model,
    and reports configurations, and uses these functions to generate the
    hydra-zen configuration.

    Returns:
        Config : Type[DataClass]
        See the documentation for hydra_zen.make_config for more information.
    """
    pbuilds = make_custom_builds_fn(
        zen_partial=True, populate_full_signature=True
    )

    def create_dataset_config(
        name, dl_root, data_file, rel_path, url, process_method, process_args
    ):
        return dict(
            data_file=data_file,
            dl_root=dl_root,
            dl_path="${.dl_root}/${.data_file}",
            rel_path=rel_path,
            url=url,
            derived=dict(
                process_method=process_method,
                process_args=process_args,
                rel_path="${data_external.processed_path}/"
                + f"{name}_processed.h5ad",
            ),
        )

    def create_model_config(
        source,
        name,
        model_suffix,
        vector_field_basis,
        **custom_training_parameters,
    ):
        return dict(
            path="${paths.models}/" + f"{name}_model{model_suffix}",
            model_path="${.path}/model",
            input_data_path="${data_external."
            + f"{source}.{name}"
            + ".derived.rel_path}",
            trained_data_path="${.path}/trained.h5ad",
            pyrovelocity_data_path="${.path}/pyrovelocity.pkl",
            metrics_path="${.path}/metrics.json",
            run_info_path="${.path}/run_info.json",
            vector_field_parameters=dict(basis=vector_field_basis),
            training_parameters=pbuilds(
                train_model,
                loss_plot_path="${..path}/loss_plot.png",
                patient_improve=0.0001,
                patient_init=45,
                **custom_training_parameters,
            ),
        )

    def create_reports_config(model_name: str, model_number: int):
        path = f"{paths['reports']}/{model_name}_model{model_number}"
        return dict(
            path=path,
            dataframe_path=f"{paths['data']}/processed/{model_name}_model{model_number}_dataframe.pkl.zst",
            shared_time_plot=f"{path}/shared_time.pdf",
            volcano_plot=f"{path}/volcano.pdf",
            rainbow_plot=f"{path}/rainbow.pdf",
            uncertainty_param_plot=f"{path}/param_uncertainties.pdf",
            vector_field_plot=f"{path}/vector_field.pdf",
            biomarker_selection_plot=f"{path}/markers_selection_scatterplot.tif",
            biomarker_phaseportrait_plot=f"{path}/markers_phaseportrait.pdf",
        )

    base = dict(log_level="INFO")

    paths = dict(
        data="data",
        models="models",
        reports="reports",
    )

    config = make_config(
        base=base,
        paths=paths,
        data_external=dict(
            root_path="${paths.data}/external",
            processed_path="${paths.data}/processed",
            sources=["simulate"],
            simulate=dict(
                download=["medium"],
                process=["medium"],
                sources=dict(
                    gcs_root_url="https://storage.googleapis.com/pyrovelocity/data"
                ),
                medium=create_dataset_config(
                    "simulated_medium",
                    dl_root="${data_external.root_path}",
                    data_file="simulated_medium.h5ad",
                    rel_path="${data_external.root_path}/simulated_medium.h5ad",
                    url="${data_external.simulate.sources.gcs_root_url}/simulated_medium.h5ad",
                    process_method="load_data",
                    process_args=dict(),
                ),
            ),
        ),
        model_training=dict(
            train=[
                "simulate_model1",
                "simulate_model2",
            ],
            simulate_model1=create_model_config(
                "simulate",
                "medium",
                1,
                "umap",
                cell_state="leiden",
                guide_type="auto_t0_constraint",
                max_epochs=1000,
            ),
            simulate_model2=create_model_config(
                "simulate",
                "medium",
                2,
                "umap",
                cell_state="leiden",
                max_epochs=1000,
                offset=True,
            ),
        ),
        reports=dict(
            model_summary=dict(
                summarize=[
                    "simulate_model1",
                    "simulate_model2",
                ],
                simulate_model1=create_reports_config("medium", 1),
                simulate_model2=create_reports_config("medium", 2),
            ),
        ),
    )
    return config


def initialize_hydra_config() -> DictConfig:
    """Initialize Hydra configuration for PyroVelocity pipeline.

    Returns:
        DictConfig: A Hydra configuration object containing dataset, model training,
        and reporting configurations for the pyrovelocity pipeline.

    Example:
        >>> cfg = initialize_hydra_config()
    """
    GlobalHydra.instance().clear()
    config = hydra_zen_configure()
    hydra.initialize(version_base="1.2", config_path=None)
    cs = ConfigStore.instance()
    cs.store(name="config", node=config)
    return hydra.compose(config_name="config")


def config_setup(config_path: str) -> Union[DictConfig, ListConfig]:
    """Convert template into concrete configuration file.
    Args:
        config_path {Text}: path to config
    """

    logger = get_pylogger(name="CONF")

    template_config_path = config_path.replace(
        "config.yaml", "template-config.yaml"
    )
    conf = OmegaConf.load(template_config_path)
    with open(config_path, "w") as conf_file:
        OmegaConf.save(config=conf, f=conf_file, resolve=True)

    conf = OmegaConf.load(config_path)
    print_config_tree(conf, logger, ())
    return conf


@rank_zero_only
def print_config_tree(
    cfg: DictConfig,
    logger,
    print_order: Sequence[str] = (
        "datamodule",
        "model",
        "callbacks",
        "logger",
        "trainer",
        "paths",
        "extras",
    ),
    resolve: bool = False,
    save_to_file: bool = False,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.

    Args:
        cfg (DictConfig): Configuration composed by Hydra.
        print_order (Sequence[str], optional): Determines in what order config components are printed.
        resolve (bool, optional): Whether to resolve reference fields of DictConfig.
        save_to_file (bool, optional): Whether to export config to the hydra output folder.
    """

    style = "dim"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    queue = []

    # add fields from `print_order` to queue
    for field in print_order:
        queue.append(field) if field in cfg else logger.warning(
            f"Field '{field}' not found in config. Skipping '{field}' config printing..."
        )

    # add all the other fields to queue (not specified in `print_order`)
    for field in cfg:
        if field not in queue:
            queue.append(field)

    # generate config tree from queue
    for field in queue:
        branch = tree.add(field, style=style, guide_style=style)

        config_group = cfg[field]
        if isinstance(config_group, DictConfig):
            branch_content = OmegaConf.to_yaml(config_group, resolve=resolve)
        else:
            branch_content = str(config_group)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    # print config tree
    rich.print(tree)

    # save config tree to file
    if save_to_file:
        with open(Path(cfg.paths.output, "config_tree.log"), "w") as file:
            rich.print(tree, file=file)


if __name__ == "__main__":
    args_parser = argparse.ArgumentParser()
    args_parser.add_argument("--config", dest="config", required=True)
    args = args_parser.parse_args()

    config_setup(config_path=args.config)
